//*******************************************************
//					Kineata Map Firmware
//*******************************************************
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "LPC214x.h"
#include "serial.h"
#include "rprintf.h"
#include "KinetaMapPins.h"
#include "spi0.h"

//*******************************************************
//				Memory Management Libraries
//*******************************************************
#include "rootdir.h"
#include "sd_raw.h"
#include "fat16.h"
#include "main_msc.h"

//*******************************************************
//					External Component Libs
//*******************************************************
#include "ADXL345.h"
#include "gps.h"

//*******************************************************
//					Core Functions
//*******************************************************
void bootUp(void);
static void timer0ISR(void);
static void RxData0ISR(void);
static void SpuriousHandlerISR(void);
void reset(void);
char string_compare(const char *search_string, const char *find_string);
void parseGGA(const char *gps_string);
void sendParsedData(void);
int get_adc_1(char channel);
void createLogFile(void);
void saveData(struct fat16_file_struct **file_name, const char * const buf, const int buf_size);
void reverse(char s[]);
void itoa(int n, char s[]);
void configBluetooth(void);
char load_settings(void);
int configMenu(void);
void initializeGPS(void);
void initializeTimer(char Frequency);
void initializeBluetooth(void);
void createNMEAFile(void);
void initializeAccelerometer(char scale);
char compareChecksum(const char *gps_string);
void parseRMC(const char *gps_string);
int parseGPS(const char *gps_string);

//*******************************************************
//					Global Variables
//*******************************************************
int baud_rate0 = 4800, baud_rate1 = 115200;
char bt_fastmode=0, gps_message_complete=0, new_gps_data=0;
int gps_message_index=0, gps_message_size=0, final_gps_message_size=0;
char final_message[GPS_BUFFER_SIZE], gps_message[GPS_BUFFER_SIZE];
GPSdata GPS;

char RTC_Set;

signed int acceleration_x, acceleration_y, acceleration_z;
char new_sensor_data=0;
char read_sensors=0;

int battery_level;

struct fat16_file_struct * LOG_FILE=NULL;
struct fat16_file_struct* fd;
int file_number;
char log_data[512], name[32];
char log_buffer[100];
int log_data_index=0;

char settingsString[100];
char config_menu_setting, gps_logging, accel_logging, gps_over_bt, batt_logging, nmea_logging=ACTIVE;
char enable_waas, full_scale_accel, enable_rmc, enable_gga, gps_message_type, accel_frequency;
int menu_option=0;
char data_led_blink=0;
char fix_led_blink=0;
char gps_data_saved=0;

//Log Parameters for logging the NMEA file
struct fat16_file_struct * NMEA_FILE=NULL; //File structure for current log file
char nmea_data[1024];//
int nmea_data_index=0;

int main (void)
{
//*******************************************************
//					Local Variables
//*******************************************************


//*******************************************************
//					Main Code
//*******************************************************
	//Initialize ARM I/O
	bootUp();
	GreenOn();
	VICIntEnClr = RxData0_int | Timer0_int;	
	
	//Get the settings from the config file
	while(!load_settings()){
		//If we got an error loading the settings, delete the settings file and try again!
		root_delete((char*)CONFIG_FILE);
		sd_raw_sync();		
	}
	GreenOff();
	BlueOn();
	
	//If the configuration menu setting is ACTIVE, wait for a bluetooth connection and then display the menu
	if(config_menu_setting==ACTIVE || gps_over_bt==ACTIVE){
		initializeBluetooth();
		if(config_menu_setting==ACTIVE){
			while((IOPIN1 & BT_CON)!=BT_CON);		//If the config menu option is active, wait until we get a BT connection
			while(((IOPIN1 & BT_CON)==BT_CON) && (config_menu_setting==ACTIVE)){
				menu_option=configMenu();
				if(menu_option=='9'){
					rprintf("\n\rStarting Logging:\n\r");
					break;
				}
			}
		}
	}

	//If no GPS messages are enabled, override the gps logging and gps_over_bt flags
	if(enable_rmc==INACTIVE && enable_gga==INACTIVE){
		gps_logging=INACTIVE;
		gps_over_bt=INACTIVE;
		nmea_logging=INACTIVE;	
	}

	//If we're not sending the GPS data over bluetooth, turn bluetooth off since we've already exited the configuration menu
	if(gps_over_bt!=ACTIVE)BluetoothOff();

	//Initialize Peripheral components and files
	//Check to see if any parameters are going to be logged; if so create a new file for them
	if(gps_logging==ACTIVE || accel_logging==ACTIVE || batt_logging == ACTIVE)createLogFile();
	//Check to see if we're going to log an NMEA file. If so, create a file for this too!
	if(nmea_logging==ACTIVE)createNMEAFile();
	
	//Only initialize the timer interrupts if we are configured to read sensors
	if(accel_logging==ACTIVE || batt_logging==ACTIVE){
		initializeTimer(accel_frequency);
	}
	if(accel_logging==ACTIVE)initializeAccelerometer(full_scale_accel);	//Initialize the Acceleromter.
	else powerdownAccel();
	
	//Initialize the GPS receiver
	if(gps_logging==ACTIVE || gps_over_bt==ACTIVE)initializeGPS();
	else{
		GPSOff();
		GPS.Fix='0';
	}	
	//Set RTC
	CCR &= ~(1<<0);	//Disable the RTC
	CCR |= (1<<1); 	//Reset the RTC
	HOUR = 0;
	MIN = 0;
	SEC = 0;
	RTC_Set=1;
	CCR &= ~(1<<1);	//Enable the RTC
	CCR |= (1<<0);		//Enable the Clock			
	
	VICIntEnable = RxData0_int | Timer0_int;	//Start UART0 interrupts
	BlueOff();
	
	//Start Listening/Logging Data
	while(1){
		//If a USB Cable gets plugged in, stop everything!
		if(IOPIN0 & (1<<23))
		{
			BlueOff();
			GreenOff();
			VICIntEnClr = RxData0_int | Timer0_int;	//Stop all running interrupts
			//Save current logged data and close the file before allowing USB communication
			powerdownAccel();
			if ( NULL != LOG_FILE ) {
				saveData(&LOG_FILE, log_data, log_data_index);
				fat16_close_file(LOG_FILE);
				log_data_index=0;
			}
			if ( NULL != NMEA_FILE ) {
				saveData(&NMEA_FILE, nmea_data, nmea_data_index);
				fat16_close_file(NMEA_FILE);
				nmea_data_index=0;
			}	
			main_msc();								//Open the mass storage device
			reset();								//Reset to check for new FW
		}
	
		//If there is a new GPS message available, copy it and set the current time
		if(gps_message_complete==1){
			VICIntEnClr = RxData0_int | Timer0_int;
			//Copy the GPS string into a buffer(final_message) so we can continue logging new GPS strings into 'gps_message'
			for(int i=0; i<gps_message_size; i++){
				final_message[i]=gps_message[i];
				gps_message[i]='\0';
			}
			final_message[gps_message_size]='\n';
			final_gps_message_size=gps_message_size+1;
			gps_message_complete=0;
			VICIntEnable = RxData0_int | Timer0_int;
			
			//Parse the GGA message. If the fix is valid, and there are no errors in the message, log the data and set the flag	
			if(parseGPS(final_message)){
				//If we were able to parse the entire gps message, than we have new gps data to log
				if(VALID_GPS)new_gps_data=1;
				
				if(nmea_logging==ACTIVE){
					//Add the gps message to the nmea buffer
					for(int i=0; i<final_gps_message_size; i++)nmea_data[nmea_data_index++]=final_message[i];
					 
					//If the nmea buffer is full then log the buffer to the NMEA file
					if(nmea_data_index >= MAX_BUFFER_SIZE){
						VICIntEnClr = RxData0_int | Timer0_int;
						saveData(&NMEA_FILE, nmea_data, nmea_data_index);
						nmea_data_index=0;
						VICIntEnable = RxData0_int | Timer0_int;
						for(int i=0; i<nmea_data_index; i++)nmea_data[i]='\0';
						nmea_data_index=0;
					}
				}
				if(VALID_GPS){
					//Set RTC
					CCR &= ~(1<<0);	//Disable the RTC
					CCR |= (1<<1); 	//Reset the RTC
					HOUR = ((GPS.Time[0]-'0')*10) + (GPS.Time[1]-'0');
					MIN = ((GPS.Time[2]-'0')*10) + (GPS.Time[3]-'0');
					SEC = ((GPS.Time[4]-'0')*10) + (GPS.Time[5] -'0');
					RTC_Set=1;
					CCR &= ~(1<<1);	//Enable the RTC
					CCR |= (1<<0);		//Enable the Clock
				}
			}
		}
		
		//Check to see if it's time to read the sensors
		if(read_sensors==1){
			VICIntEnClr |= Timer0_int;
			//Get Acceleration
			acceleration_x = accelX();
			acceleration_y = accelY();
			acceleration_z = accelZ();
			
			//Get Battery Voltage
			battery_level = ((get_adc_1(0x80)*3300)/1024)*2;	//Report Battery voltage in mV. from AD1.4(P0.13)
			
			new_sensor_data=1;	//Set the flag to tell the logging routine that there is new sensor data to be saved
			read_sensors=0;		//Reset the "read sensor" flag

			VICIntEnable |= Timer0_int;
		}		
		
		//Check to see if it's time to log an acceleration value.
		if(new_gps_data || new_sensor_data){
			//Log Time
			//If there is GPS data, use this time,
			if(new_gps_data)for(int i=0; i<10; i++){
				log_data[log_data_index++]=GPS.Time[i];
			}	
			//If there is not GPS data, use the RTC time
			else if(RTC_Set){
					log_data[log_data_index++]=(HOUR / 10) + '0';
					log_data[log_data_index++]=(HOUR % 10) + '0';
					log_data[log_data_index++]=(MIN / 10) + '0';
					log_data[log_data_index++]=(MIN % 10) + '0';
					log_data[log_data_index++]=(SEC / 10) + '0';
					log_data[log_data_index++]=(SEC % 10) + '0';
			}
			log_data[log_data_index++]=',';
			
			if(accel_logging==ACTIVE && new_sensor_data){
				//Log Acceleration
				log_data_index += (int) sprintf(log_data+log_data_index, "%d,%d,%d,", acceleration_x, acceleration_y, acceleration_z);
				acceleration_x=0;
				acceleration_y=0;
				acceleration_z=0;
			}
			else for(int i=0; i<3; i++)log_data[log_data_index++]=',';
			
			if(batt_logging==ACTIVE && new_sensor_data){
				//Log Battery Level
				itoa(battery_level, log_buffer);
				for(int i=0; log_buffer[i] != '\0'; i++)log_data[log_data_index++] = log_buffer[i];
				log_data[log_data_index++]=',';
				battery_level=0;
			}
			else log_data[log_data_index++]=',';
			
			//If we have valid gps data, add it to the log buffer and then save it
			if(new_gps_data){
				gps_data_saved=1;
				log_data[log_data_index++]=GPS.Fix;
				log_data[log_data_index++]=',';
				GPS.Fix='0';
				
				for(int i=0; i<9; i++)log_data[log_data_index++]=GPS.Latitude.position[i];
				log_data[log_data_index++]=',';
				
				log_data[log_data_index++]=GPS.Latitude.direction;
				log_data[log_data_index++]=',';
				
				for(int i=0; i<10; i++)log_data[log_data_index++]=GPS.Longitude.position[i];
				log_data[log_data_index++]=',';

				log_data[log_data_index++]=GPS.Longitude.direction;
				log_data[log_data_index++]=',';
				

				if(gps_message_type==GGA){
					for(int i=0; i<6; i++)log_data[log_data_index++]=GPS.Altitude[i];
					log_data_index += (int)sprintf(log_data+log_data_index, ",GGA");
				}
				else if(gps_message_type==RMC)log_data_index += (int)sprintf(log_data+log_data_index, ",RMC");
				else log_data[log_data_index++]=',';

			}
			else for(int i=0; i<6; i++)log_data[log_data_index++]=',';
			log_data[log_data_index++]='\n';
				
			//Only Save Data if the buffer is full!
			if(log_data_index >= MAX_BUFFER_SIZE){
				VICIntEnClr = RxData0_int | Timer0_int;
				saveData(&LOG_FILE, log_data, log_data_index);
				log_data_index=0;
				VICIntEnable = RxData0_int | Timer0_int;
				
				//Set the status indicator based on the validity of the GPS messages and whether or not data has been saved
				if(gps_data_saved==1 && fix_led_blink == 0)fix_led_blink=accel_frequency;
				else if(data_led_blink ==0)data_led_blink=accel_frequency;					
				gps_data_saved=0;
			}				
	
			new_gps_data=0;
			new_sensor_data=0;
		}
	}	
	
    return 0;
}


//Usage: delay_ms(1000);
//Inputs: int count: Number of milliseconds to delay
//The function will cause the firmware to delay for "count" milleseconds.
void delay_ms(int count)
{
    int i;
    count *= 10000;
    for (i = 0; i < count; i++)
        asm volatile ("nop");
}

//Usage: bootUp();
//Inputs: None
//This function initializes the serial port, the SD card, the I/O pins and the interrupts
void bootUp(void)
{
	//Initialize UART for RPRINTF
    rprintf_devopen(putc_serial0); //Init rprintf
	init_serial0(baud_rate0);
    delay_ms(10); //Delay for power to stablize
	
    //Bring up SD and FAT
    if(!sd_raw_init())
    {
        rprintf("SD Init Error\n");
    }
    if(openroot())
    {
        rprintf("SD OpenRoot Error\n");
    }
	
	//Initialize the SPI bus
	SPI0_Init();
	//Setup the SPI Port
    S0SPCCR = 64;	// SCK = 1 MHz, counter > 8 and even
    S0SPCR  = 0x20; // Master, no interrupt enable, 8 bits	
	
	//Enable AD conversion on P0.22(AD1.7) FOR VLIPO_MEAS
	PINSEL1 &= ~(3<<12);
	PINSEL1 |= (1<<12);

	//Disable USB LED functionality
	PINSEL1 &= ~(3<<31);
	
	//Initialize I/O Ports and Peripherals
	IODIR0 = BT_RTS | SSEL1 | BT_EN | ACCEL_CS1;
	IODIR0 &= ~(BT_CTS | VLIPO_MEAS | ACCEL_INT1 | ACCEL_INT2);
	
	IODIR1 = STAT1 | FLASH_CS | CD1 | CD2 | GPS_EN | STAT0;
	IODIR1 &= ~BT_CON;

	//Turn Off LED's
	GreenOff();
	BlueOff();
	
	//Make sure Accelerometer is NOT selected
	UnselectAccel();

    //Setup the Interrupts
	//Enable Interrupts
	VPBDIV=1;										// Set PCLK equal to the System Clock	
	VICIntSelect = ~(RxData0_int | Timer0_int);
	VICVectCntl0 = 0x26;
	VICVectAddr0 = (unsigned int)RxData0ISR;
	VICVectCntl1 = 0x24;
	VICVectAddr1 = (unsigned int)timer0ISR;
	
	VICDefVectAddr = (unsigned int)SpuriousHandlerISR;
	
	//Setup the UART0 Interrupt
	U0IER = 0x01;				//Enable FIFO on UART with RDA interrupt (Receive Data Available)
	U0FCR &= 0x3F;				//Enable FIFO, set RDA interrupt for 1 character
	
	//Setup the RTC
	CCR = (1<<4);		//Select external 32.768 KHz clock for RTC(Keep RTC Disabled for now!)
}

//Usage: None (Automatically Called by FW)
//Inputs: None
//This function is a global interrupt called by a match on the Timer 1 match.  The interrupt
// is responsible for determining if a button has been pressed or if the screen has been rotated
// and setting the appropriate global flag if either has occured.
static void timer0ISR(void)
{
	//Interrupt Code Here
	read_sensors=1;

	if(data_led_blink>0){
		if(data_led_blink/10==(accel_frequency/10))BlueOn();
		else BlueOff();
		data_led_blink--;
	}
	else BlueOff();
	
	if(fix_led_blink>0){
		if(fix_led_blink/10==(accel_frequency/10))GreenOn();
		else GreenOff();
		fix_led_blink--;
	}	
	else GreenOff();
	
	T0IR = 0xFF;						//Clear the timer interrupt
	VICVectAddr =0;						//Update the VIC priorities
}

//Usage: None (Automatically Called by FW)
//Inputs: None
static void SpuriousHandlerISR(void){
	uint32_t trash=0;

	trash=U0IIR;		//Clear the Rx Interrupt without ruining any characters in the Rx Buffer
	T0IR = 0xFF;		//Clear the timer interrupt
	VICVectAddr = 0;		//Update the VIC priorities	
}

//Usage: None (Automatically Called by FW)
//Inputs: None
static void RxData0ISR(void)
{
	char val = (char)U0RBR;
	if(gps_over_bt==ACTIVE)putc_serial1(val);	//If configured to send gps over bluetooth, put recvd char on UART1
	//When we get a character on UART0, save it to the GPS message buffer
	if(val=='\n' || gps_message_index>GPS_BUFFER_SIZE){ 	//Newline means the current message is complete
		//gps_message[gps_message_index]= val;	//Copy the newline character to the buffer to indicate the end of the message
		gps_message_complete=1;					//Set a flag for the main FW
		//gps_message_size=gps_message_index+1;
		gps_message_size=gps_message_index;
		gps_message_index=0;
	}
	else{
		//If we get the start character, reset the index
		if(val == '$')gps_message_index=0;
		gps_message[gps_message_index++]= val;
		gps_message_complete=0;
	}
	VICVectAddr =0;						//Update the VIC priorities
}

void reset(void)
{
    // Intentionally fault Watchdog to trigger a reset condition
    WDMOD |= 3;
    WDFEED = 0xAA;
    WDFEED = 0x55;
    WDFEED = 0xAA;
    WDFEED = 0x00;
}

//Use to find a string within a search string
//search : "hithere\0"
//find : "the\0" - return OK
//***************************** UNTESTED *********************************
char string_compare(const char *search_string, const char *find_string)
{

    int find_spot, search_spot;
    char spot_character, search_character;
    find_spot = 0;

    for(search_spot = 0 ; ; search_spot++)
    {
        if(find_string[find_spot] == '\0') return OK; //We've reached the end of the search string - that's good!
        if(search_string[search_spot] == '\0') return ERROR; //End of string found

        spot_character = find_string[find_spot]; //Compiler limit
        search_character = search_string[search_spot]; //Compiler limit

        if(spot_character == search_character) //We found another character
            find_spot++; //Look for the next spot in the search string
        else if(find_spot > 0) //No character found, so reset the find_spot
            find_spot = 0;
    }

    return 0;
}

//This functions splits a GGA message into the
//portions and assigns them to components of
//a GPS structure
void parseGGA(const char *gps_string){
	int i=0;

	//Parse the GGA Message.  1st portion dismissed
	while(gps_string[i] != ',')i++;
	i++;
	//Second portion is UTC timestamp
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Time[j]=gps_string[i];
		i++;
	}
	i++;
	//Third portion is Latitude
	for(int j=0;gps_string[i] != ',';j++){
		GPS.Latitude.position[j]=gps_string[i];
		i++;
	}
	i++;			
	//Fourth portion is Latitude direction
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Latitude.direction=gps_string[i];
		i++;
	}
	i++;	
	//Fifth portion is Long.
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Longitude.position[j]=gps_string[i];
		i++;
	}
	i++;			
	//Sixth portion is Long direction
	while(gps_string[i] != ','){
		GPS.Longitude.direction=gps_string[i];
		i++;
	}
	i++;		
	//Seventh portion is fix
	while(gps_string[i] != ','){
		GPS.Fix=gps_string[i];
		i++;
	}
	i++;
	//8th portion dismissed
	while(gps_string[i] != ',')i++;
	i++;				
	//8th portion dismissed
	while(gps_string[i] != ',')i++;
	i++;				
	//10th portion is Altitude
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Altitude[j]=gps_string[i];
		i++;
	}
}

void sendParsedData(void){
	rprintf("Time: ");
	for(int j=0; j<6; j++){
		if(j==2)rprintf(":");
		if(j==4)rprintf(".");
		rprintf("%c", GPS.Time[j]);
	}
	rprintf("\n\r");
	rprintf("Latitude: ");
	for(int j=0; j<15; j++)rprintf("%c", GPS.Latitude.position[j]);
	rprintf("%c\n\r", GPS.Latitude.direction);
	rprintf("Longitude: ");
	for(int j=0; j<15; j++)rprintf("%c", GPS.Longitude.position[j]);
	rprintf("%c\n\r", GPS.Longitude.direction);
	rprintf("Altitude: ");
	for(int j=0; j<10; j++)rprintf("%c", GPS.Altitude[j]);
	rprintf("\n\r");
	rprintf("Fix: %c\n\r", GPS.Fix);
}

//Usage: accel = get_adc_1(CHANNEL);
//Inputs: int channel - integer corresponding to the ADC channel to be converted
//Outputs: None
//Description: Returns the raw analog to digital conversion of the input channel.  
int get_adc_1(char channel)
{
    int val;
    AD1CR = 0;
    AD1GDR = 0;

	AD1CR = 0x00200E00 | channel;
    AD1CR |= 0x01000000;
    do
    {
        val = AD1GDR;                   // Read A/D Data Register
    }
    while ((val & 0x80000000) == 0);  //Wait for the conversion to complete
    val = ((val >> 6) & 0x03FF);  //Extract the A/D result

    return val;
}

void createLogFile(void){

//Create the NMEA Log File	
	//Set an initial file name
	sprintf(name, "KinetaMap%03d.csv", file_number);
	//Check to see if the file already exists in the root directory.
    while(root_file_exists(name))
    {
        file_number++;	//If the file already exists, increment the file number and check again.
        if(file_number == 250)
        {
            rprintf("\nToo many files in root!\n");
        }
        sprintf(name, "KinetaMap%03d.csv", file_number);
    }
	//Get the file handle of the new file.  We will log the data to this file
    LOG_FILE = root_open_new(name);
	//Now that we have the file opened, let's put a label in the first row
	fat16_write_file(LOG_FILE, (const unsigned char *)LOG_FILE_HEADER, strlen(LOG_FILE_HEADER));
	sd_raw_sync();		

}

void createNMEAFile(void){
//Create the NMEA Log File	
	//Set an initial file name
	sprintf(name, "KinetaMapNMEA%03d.csv", file_number);
	//Check to see if the file already exists in the root directory.
    while(root_file_exists(name))
    {
        file_number++;	//If the file already exists, increment the file number and check again.
        if(file_number == 250)
        {
            //rprintf("\nToo many files in root!\n");
        }
        sprintf(name, "KinetaMapNMEA%03d.csv", file_number);
    }
    //Get the file handle of the new file.  We will log the data to this file
	NMEA_FILE = root_open_new(name);
	//Now that we have the file opened, let's put a label in the first row
	//fat16_write_file(NMEA_FILE, (const unsigned char *)NMEA_FILE_HEADER, strlen(NMEA_FILE_HEADER));
	sd_raw_sync();
}

//Usage: saveData(FILE_NAME, log_data, log_data_index);
//Inputs: char *buf - character array to be saved
//		  int buf_size - size of character array
//Output: buffer array is saved to LOG_FILE
//Description: Saves the buf character array to the SD card.
//CONDITIONS: LOG_FILE must be initialized to the handle of an open file.
void saveData(struct fat16_file_struct **file_name, const char * const buf, const int buf_size)
{
	int error=0;
	
	if((buf_size > 0) && (*file_name != NULL)){
		//Try writing the data to the card up to 10 times.
		while(error<10){
			if(fat16_write_file(*file_name, (const unsigned char*)buf, buf_size) < 0)error+=1;
			else break;
			delay_ms(100);
		}
		//If we've tried writing the data 10 times and still haven't succeeded, reset the device.
		if(error==10)reset();
		
		error=0;
		//Try syncing the card up to 10 times
		while(error<10){
			if(!sd_raw_sync())error+=1;
			else break;
			delay_ms(100);
		}
		//If we've tried syncing 10 times and still haven't succeeded, reset the device
		if(error==10)reset();
	}
}

/* itoa:  convert n to characters in s */
void itoa(int n, char s[])
{
    int i=0;

    do {       /* generate digits in reverse order */
        s[i++] = n % 10 + '0';   /* get next digit */
    } while ((n /= 10) > 0);     /* delete it */

    s[i] = '\0';
    reverse(s);
} 

/* reverse:  reverse string s in place */
void reverse(char s[])
{
    int c, i, j;

    for (i = 0, j = strlen(s)-1; i<j; i++, j--) {
        c = s[i];
        s[i] = s[j];
        s[j] = c;
    }
}

void configBluetooth(void){
	//Enter Command Mode
	delay_ms(100);
	rprintf("$$$");
	delay_ms(100);

	//Disable the Config Timer
	rprintf("ST,0\r");
	delay_ms(50);
	
	//Activate Fast Data Mode
	rprintf("SN,KinetaMap\r");
	delay_ms(50);		
	
	//Activate Fast Data Mode
	rprintf("F,1\r");
	delay_ms(50);
	
	//Exit Command Mode
	//rprintf("---");
	//delay_ms(50);
}

//Loads parameters from the configuration file.
char load_settings(void){
	struct fat16_file_struct * settingsFile;
	unsigned char current_character;
	unsigned char settings_index=0;
	
	//If the config file doesn't exist, then create it!
	if(!root_file_exists(CONFIG_FILE)){
		LOG_FILE = root_open_new(CONFIG_FILE);	
		fat16_write_file(LOG_FILE, (unsigned char*)CONFIG_STRING, strlen(CONFIG_STRING));
		sd_raw_sync();	
		fat16_close_file(LOG_FILE);
	}	
	
	//Open the configuration File
	settingsFile=root_open(CONFIG_FILE);
	//Get the contents of the configuration file into the "settingsString"
	if(settingsFile != NULL){
		int i=0;
		while(fat16_read_file(settingsFile, &current_character, 1)>0){
			settingsString[i++]=current_character;	//Read the next char in the file
			if(i>MAX_SETTINGS_FILE_SIZE)break;	//If there are too many characters, get out of the loop!
		}
	}
	//settingsString is now filled with configuration settings.
	//See KinetaMapPins.h for description of Configuration File	
	fat16_close_file(settingsFile);
	
	//Traverse through the setting file until we see an equals sign(=)
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//We've found the first '=', so now get the first setting
	if(isdigit(settingsString[settings_index+1]))config_menu_setting=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the GPS Logging setting
	if(isdigit(settingsString[settings_index+1]))gps_logging=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the Accel Logging setting
	if(isdigit(settingsString[settings_index+1]))accel_logging=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the GPS Over Bluetooth setting
	if(isdigit(settingsString[settings_index+1]))gps_over_bt=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the Battery Logging setting
	if(isdigit(settingsString[settings_index+1]))batt_logging=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the NMEA Logging setting
	if(isdigit(settingsString[settings_index+1]))nmea_logging=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the WAAS setting
	if(isdigit(settingsString[settings_index+1]))enable_waas=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the Accelerometer Scale setting
	if(isdigit(settingsString[settings_index+1]))full_scale_accel=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;	
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the RMC GPS Messages setting
	if(isdigit(settingsString[settings_index+1]))enable_rmc=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;

	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the GGA GPS Messages setting
	if(isdigit(settingsString[settings_index+1]))enable_gga=settingsString[++settings_index]-'0';
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	else return 0;
	if(settingsString[settings_index+1]!='\r')return 0;
	
	//Navigate to the next setting
	while(settingsString[settings_index]!='='){
		if(settingsString[settings_index]=='\0')return 0;
		settings_index+=1;
	}
	//Load the Accelerometer Frequency setting
	accel_frequency=0;
	while(settingsString[settings_index]!='\r'){
		if(isdigit(settingsString[settings_index+1]))accel_frequency=(accel_frequency*10)+(settingsString[++settings_index]-'0');
		else if(settingsString[++settings_index]!='\r')return 0;
		//else settings_index+=1;
	}
	//If we don't see a valid setting, exit the function because we have an invalid settings file
	//if(settingsString[settings_index+1]!='\r')return 0;	
	
	//Now we've loaded all of the settings!
	return 1;
	
}

int configMenu(void){
	int selection=0, error_code=1;
	char character, tempbuffer[50];
	
	rprintf("----------- KinetaMap Config Menu -----------\n");
	rprintf("                FW Version 1.2\n");
	rprintf("              SparkFun Electronics\n");
	rprintf("_____________________________________________\n");
	rprintf("1) Read directory\n");
	rprintf("2) Read file\n");
	rprintf("3) Delete file\n");
	rprintf("4) Format SD card\n");
	rprintf("5) Listen to GPS (press any key to stop)\n");
	rprintf("6) Check Accelerometer(press any key to stop)\n");
	rprintf("7) Change settings in the config file\n");
	
	rprintf("9) Exit menu and start logging\n");

	rprintf("\n\n\rCurrent Settings:\n\r");
	rprintf("Menu: %d\n\r", config_menu_setting);
	rprintf("GPS Logging: %d\n\r", gps_logging);
	rprintf("Accel Logging: %d\n\r", accel_logging);
	rprintf("GPS over BT: %d\n\r", gps_over_bt);
	rprintf("Battery Logging: %d\n\r", batt_logging);
	rprintf("NMEA Logging: %d\n\r", nmea_logging);
	rprintf("WAAS Enabled: %d\n\r", enable_waas);
	rprintf("Accel Range: %d\n\r", full_scale_accel);
	rprintf("RMC Messages: %d\n\r", enable_rmc);
	rprintf("GGA Messages: %d\n\r", enable_gga);
	rprintf("Accelerometer Frequency: %d\n\r", accel_frequency);

	selection=getc1();
	rprintf("\nOption = %c\n\n", selection);
	
	if(selection=='1'){
		rootDirectory_files_stream(1);
		rprintf("Files in Root Directory:\n");
		character = 1;
		while (character != 0)
		{
			character = rootDirectory_files_stream(0);
			
			if (character == ',')
			{
				putc_serial1('\n');
				putc_serial1('\r');
			}
			else putc_serial1(character);
		}
		rprintf("\n");
		character = 0;	
	}
	if(selection=='2'){
		for (int m = 0; m < 50; m++) tempbuffer[m] = 0;
		rprintf("Enter the name of the file you wish to read: ");
		for (int m = 0; m < 50; m++)
		{
			character = getc1();
			
			if ((character == 10) | (character == 13)) break;
			putc_serial1(character);
			tempbuffer[m] = character;
		}
		
		fd = root_open((char*)tempbuffer);
		if (fd == 0)
		{
			rprintf("\nFailed to open file.\n\n");
			return 0;
		}
		
		rprintf("\n\nFile Contents:\n");
		while(error_code > 0)
		{
			tempbuffer[0] = 0;
			
			error_code=fat16_read_file(fd, (uint8_t*)tempbuffer, 1);
			if(error_code<0){
				rprintf("\n\nFailed to read %d\n\n",character);
				return 0;
			}
			
			if (tempbuffer[0] != 0) putc_serial1(tempbuffer[0]);
			if(tempbuffer[0]=='\n')putc_serial1('\r');			
		}
		
		fat16_close_file(fd);
		sd_raw_sync();
		
		rprintf("\nEnd.\n\n");	
	}
	if(selection=='3'){
		for (int m = 0; m < 50; m++) tempbuffer[m] = 0;
		rprintf("Enter the name of the file you wish to delete: ");
		for (int m = 0; m < 50; m++)
		{
			character = getc1();
			
			if ((character == 10) | (character == 13)) break;
			putc_serial1(character);
			tempbuffer[m] = character;
			
		}
		
		character = root_delete((char*)tempbuffer);
		if (character == 0) rprintf("\nFile Deleted.\n\n");
		else rprintf("\nFailed.\n\n");
		
		sd_raw_sync();	
	}
	if(selection=='4'){
		root_format();
		rprintf("Card Formatted\n\n");
	}
	if(selection=='5'){
		initializeGPS();
		delay_ms(500);
		//Keep printing the character received on UART0 to UART1 until a character is received on UART1
		while(!(U1LSR & (1<<0)))	//If a character is received on UART1(input from Bluetooth), stop sending the GPS messages to the Bluetooth module
		  if(U0LSR & (1<<0))		//If nothings been seen on UART1, check to see if we've received anything on UART0
			putc_serial1(U0RBR);	//If we got something on UART0; print it out on UART1(Send GPS data to Bluetooth)
		rprintf("\n");		
	}
	if(selection=='6'){
		initializeAccelerometer(full_scale_accel);
		while(!(U1LSR & (1<<0))){
			rprintf("X=%d\tY=%d\tZ=%d\n", (signed char)accelX(),(signed char)accelY(),(signed char)accelZ());
			delay_ms(200);
		}
		rprintf("\n\n");
	}
	if(selection=='7'){
		char option;
		
		if(root_file_exists(CONFIG_FILE))root_delete(CONFIG_FILE);
		rprintf("\nConfiguration File Deleted!\n\n");
		
		rprintf("Enable Configuration Menu(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')config_menu_setting=ACTIVE;
		else config_menu_setting=INACTIVE;
		
		rprintf("Log GPS Data(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')gps_logging=ACTIVE;
		else gps_logging=INACTIVE;	
		
		rprintf("Log Accelerometer Data(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')accel_logging=ACTIVE;
		else accel_logging=INACTIVE;		
		
		rprintf("Log Battery Level(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')batt_logging=ACTIVE;
		else batt_logging=INACTIVE;
		
		rprintf("Send GPS data over Bluetooth(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')gps_over_bt=ACTIVE;
		else gps_over_bt=INACTIVE;
		
		rprintf("Create/Log NMEA strings(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')nmea_logging=ACTIVE;
		else nmea_logging=INACTIVE;			

		rprintf("Enable WAAS(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')enable_waas=ACTIVE;
		else enable_waas=INACTIVE;	

		rprintf("Use the Full Scale accelerometer mode?(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')full_scale_accel=ACTIVE;
		else full_scale_accel=INACTIVE;	
		
		rprintf("Enable RMC GPS Messages?(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')enable_rmc=ACTIVE;
		else enable_rmc=INACTIVE;	

		rprintf("Enable GGA GPS Messages?(y/n)?\n");
		option=getc1();
		rprintf("%c\n\n", option);
		if(option=='y'||option=='Y')enable_gga=ACTIVE;
		else enable_gga=INACTIVE;

		rprintf("Enter the Accelerometer Frequency in Hz (Minimum 1, Maximum 100): ");
		accel_frequency=0;
		option=getc1();
		while(option != '\n' && option != '\r'){
			if(isdigit(option))accel_frequency=(accel_frequency*10)+(option-'0');
			else{
				accel_frequency=10;
				break;
			}
			rprintf("%c", option);
			option=getc1();
		}				

		LOG_FILE = root_open_new(CONFIG_FILE);
		sprintf(settingsString, "config_menu=%c\r\nlog_gps=%c\r\nlog_accel=%c\r\nsend_gps_to_bt=%c\r\nlog_batt=%c\r\nnmea_logging=%c\r\nenable_waas=%c\r\nfull_scale_accel=%c\r\nenable_rmc=%c\r\nenable_gga=%c\r\naccel_frequency(hz)=%d\r\n",
			config_menu_setting+'0', gps_logging+'0', accel_logging+'0', gps_over_bt+'0', batt_logging+'0', nmea_logging+'0',enable_waas+'0', full_scale_accel+'0', enable_rmc+'0', enable_gga+'0', accel_frequency);
		fat16_write_file(LOG_FILE, (unsigned char *)settingsString, strlen(settingsString));
		sd_raw_sync();	
		fat16_close_file(LOG_FILE);
		
	}	
	if(selection=='9')return selection;
	return 0;
}

void initializeGPS(void){
		GPSOn();	//Turn Power on the the GPS module
		rprintf_devopen(putc_serial0); //Init rprintf for GPS UART
		init_serial0(baud_rate0);		
		delay_ms(1000);		//Wait for stabilization
		disable_all_gps_msgs();	//Turn off the GPS messages
		delay_ms(10);
		
		if(enable_gga==ACTIVE){
			enable_gps_gga_msgs(1);	//Enable 1 hz GGA messages
			delay_ms(10);
		}
		if(enable_rmc==ACTIVE){
			enable_gps_rmc_msgs(1);	//Enable 1 hz RMC messages
			delay_ms(10);
		}
		
		if(enable_waas==ACTIVE){
			enable_gps_waas();
			delay_ms(10);
		}
		else disable_waas();
		
		rprintf_devopen(putc_serial1); 	//Init rprintf on UART1 (Bluetooth Channel)
		init_serial1(baud_rate1);		//Set bluetooth Baud Rate
		
}

void initializeTimer(char Frequency){
	//Setup the Timer0 Interrupt
	T0TC = 0; // clear the timer count

	//Setupt the Timer0 Interrupt
	T0PR = 600;										//Divide Clock(60Mhz) by 600 for 100kHz PS
	T0TCR |=0X01;									//Enable the clock
	T0CTCR=0;										//Timer Mode
	T0MCR=0x0003;	

    //Set the Match Register Counter to the configured value
    //When this value will match the Counter value the actions configured in T0MCR
    //will be triggered
	if ( 0 < Frequency) {
  	    T0MR0=(100000/Frequency);	//Divide Timer(100Khz) by Frequency for ints on Frequency Hz
    } else {
	    T0MR0=10000; //Divide Timer(100Khz) by 10000 for ints on 10Hz
	}
}

void initializeBluetooth(void){
    rprintf_devopen(putc_serial1); 	//Init rprintf on UART1 (Bluetooth Channel)
	init_serial1(baud_rate1);		//Set bluetooth Baud Rate

	BluetoothOn();			//Turn on the Bluetooth device
	delay_ms(200);
	configBluetooth();		//Send configuration strings
}

void initializeAccelerometer(char scale){
	initAccel();

	//If the scale is set, we want the accelerometer to operate in 8g mode. 
	if(scale) adxl345_write(DATA_FORMAT, RANGE_1);	//Configure the Accelerometer for +/-8g;
	//If scale isn't set, then we want partial scale mode. Clear the bit in the register
	else adxl345_write(DATA_FORMAT, 0x00);	//Configure the Accelerometer for +/-8g;
		
}

char compareChecksum(const char *gps_string){
	int cur_index=0, i=0;
	char string[GPS_BUFFER_SIZE];
	char checksum_string[2];
	char checksum=0, checksum_compare_value=0;
		
	//Get the message from the GPS string
	for(i=0; gps_string[i]!='*'; i++){
		//If we reach the end of the string without seeing a checksum indicator, return an error
		if(gps_string[i]=='\n' || gps_string[i]=='\r')return 0;
		if(gps_string[i]!='$'){
			string[cur_index++]=gps_string[i];
		}
	}
	string[cur_index++]='\0';
	
	//Get the checksum value from the gps string
	for(int j=0; j<2; j++){
		//If the value of the checksum character isn't alphanumeric, return an error.
		checksum_string[j]=gps_string[++i]; //Skip the '*'
	}
	
	//Combine the two checksum characters
	checksum=((checksum_string[0]<<4)&0xF0)|(checksum_string[1]&0x0F);
	//Find the checksum of the received string
	GPS_CHECKSUM(string,checksum_compare_value)
		
	//rprintf("\n\rReceived: %x\n\rCalculated: %x\n\n\r", checksum, checksum_compare_value);
	//Compare the calculated checksum to the received checksum; if they don't match we're out!
	if(checksum != checksum_compare_value)return 0;
	return 1;
}

//Usage: parseRMC(final_message);
//Inputs: const char *gps_string - RMC NMEA string
//This functions splits a GGA message into the
//portions and assigns them to components of
//a GPS structure
//This functions splits a RMC message into the
//portions and assigns them to components of
//a GPS structure
void parseRMC(const char *gps_string){
	int i=0;
	
	//Parse the GGA Message.  1st portion dismissed
	while(gps_string[i] != ',')i++;
	i++;
	
	//Second portion is UTC timestamp
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Time[j]=gps_string[i];
		i++;
	}
	i++;
	
	//Third portion is fix
	while(gps_string[i] != ','){
		GPS.Fix=gps_string[i];
		i++;
	}	
	i++;
	
	//Fourth portion is Latitude
	for(int j=0;gps_string[i] != ',';j++){
		GPS.Latitude.position[j]=gps_string[i];
		i++;
	}
	i++;	
	
	//Fifth portion is Latitude direction
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Latitude.direction=gps_string[i];
		i++;
	}
	i++;
	
	//Sixth portion is Long.
	for(int j=0;gps_string[i] != ','; j++){
		GPS.Longitude.position[j]=gps_string[i];
		i++;
	}
	i++;			
	
	//Seventh portion is Long direction
	while(gps_string[i] != ','){
		GPS.Longitude.direction=gps_string[i];
		i++;
	}
	i++;
	
	//8th portion dismissed
	while(gps_string[i] != ',')i++;
	i++;				
	//9th portion dismissed
	while(gps_string[i] != ',')i++;
	i++;				
	//10th portion is Date
	for(int j=0;gps_string[i] != ','; j++){
		//GPS.Date[j]=gps_string[i];
		i++;
	}
}

//RETURN: 0-Error in message. Either bad checksum or insuffecient data
	   // 1-Valid GPS message
int parseGPS(const char *gps_string){
	//Make sure we have a valid string!
	if(!compareChecksum(gps_string))return 0;
	//If we didn't receive the correct SiRF header, the return an error
	if(gps_string[3] == 'G' && gps_string[4] == 'G' && gps_string[5] == 'A'){
		parseGGA(gps_string);
		gps_message_type=GGA;
		return 1;
	}
	if(gps_string[3] == 'R' && gps_string[4] == 'M' && gps_string[5] == 'C'){
		parseRMC(gps_string);
		gps_message_type=RMC;
		return 1;
	}
	gps_message_type=0;
	return 0;
	
}
